@@ -1,55 +1,17 @@
-/* ====================================================================
- * The Apache Software License, Version 1.1
+/* Copyright 2000-2005 The Apache Software Foundation or its licensors, as
+ * applicable.
  *
- * Copyright (c) 2000-2003 The Apache Software Foundation.  All rights
- * reserved.
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
+ *     http://www.apache.org/licenses/LICENSE-2.0
  *
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- *
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in
- *    the documentation and/or other materials provided with the
- *    distribution.
- *
- * 3. The end-user documentation included with the redistribution,
- *    if any, must include the following acknowledgment:
- *       "This product includes software developed by the
- *        Apache Software Foundation (http://www.apache.org/)."
- *    Alternately, this acknowledgment may appear in the software itself,
- *    if and wherever such third-party acknowledgments normally appear.
- *
- * 4. The names "Apache" and "Apache Software Foundation" must
- *    not be used to endorse or promote products derived from this
- *    software without prior written permission. For written
- *    permission, please contact apache@apache.org.
- *
- * 5. Products derived from this software may not be called "Apache",
- *    nor may "Apache" appear in their name, without prior written
- *    permission of the Apache Software Foundation.
- *
- * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
- * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
- * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
- * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
- * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
- * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
- * ====================================================================
- *
- * This software consists of voluntary contributions made by many
- * individuals on behalf of the Apache Software Foundation.  For more
- * information on the Apache Software Foundation, please see
- * <http://www.apache.org/>.
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
 
 #include "apr.h"
@@ -62,6 +24,10 @@ APR_DECLARE(apr_status_t) apr_proc_mutex_destroy(apr_proc_mutex_t *mutex)
     return apr_pool_cleanup_run(mutex->pool, mutex, apr_proc_mutex_cleanup);
 }
 
+static apr_status_t proc_mutex_no_tryacquire(apr_proc_mutex_t *new_mutex)
+{
+    return APR_ENOTIMPL;
+}
 
 #if APR_HAS_POSIXSEM_SERIALIZE
 
@@ -165,10 +131,10 @@ static apr_status_t proc_mutex_posix_release(apr_proc_mutex_t *mutex)
 {
     int rc;
 
+    mutex->curr_locked = 0;
     if ((rc = sem_post((sem_t *)mutex->interproc->filedes)) < 0) {
         return errno;
     }
-    mutex->curr_locked = 0;
     return APR_SUCCESS;
 }
 
@@ -188,7 +154,7 @@ const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_posix_methods =
 #endif
     proc_mutex_posix_create,
     proc_mutex_posix_acquire,
-    NULL, /* no tryacquire */
+    proc_mutex_no_tryacquire,
     proc_mutex_posix_release,
     proc_mutex_posix_cleanup,
     proc_mutex_posix_child_init,
@@ -269,13 +235,13 @@ static apr_status_t proc_mutex_sysv_release(apr_proc_mutex_t *mutex)
 {
     int rc;
 
+    mutex->curr_locked = 0;
     do {
         rc = semop(mutex->interproc->filedes, &proc_mutex_op_off, 1);
     } while (rc < 0 && errno == EINTR);
     if (rc < 0) {
         return errno;
     }
-    mutex->curr_locked = 0;
     return APR_SUCCESS;
 }
 
@@ -293,7 +259,7 @@ const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_sysv_methods =
 #endif
     proc_mutex_sysv_create,
     proc_mutex_sysv_acquire,
-    NULL, /* no tryacquire */
+    proc_mutex_no_tryacquire,
     proc_mutex_sysv_release,
     proc_mutex_sysv_cleanup,
     proc_mutex_sysv_child_init,
@@ -363,7 +329,7 @@ static apr_status_t proc_mutex_proc_pthread_create(apr_proc_mutex_t *new_mutex,
         return rv;
     }
 
-#ifdef HAVE_PTHREAD_MUTEXATTR_SETROBUST_NP
+#ifdef HAVE_PTHREAD_MUTEX_ROBUST
     if ((rv = pthread_mutexattr_setrobust_np(&mattr, 
                                                PTHREAD_MUTEX_ROBUST_NP))) {
 #ifdef PTHREAD_SETS_ERRNO
@@ -379,7 +345,7 @@ static apr_status_t proc_mutex_proc_pthread_create(apr_proc_mutex_t *new_mutex,
         proc_mutex_proc_pthread_cleanup(new_mutex);
         return rv;
     }
-#endif
+#endif /* HAVE_PTHREAD_MUTEX_ROBUST */
 
     if ((rv = pthread_mutex_init(new_mutex->pthread_interproc, &mattr))) {
 #ifdef PTHREAD_SETS_ERRNO
@@ -413,7 +379,7 @@ static apr_status_t proc_mutex_proc_pthread_acquire(apr_proc_mutex_t *mutex)
 #ifdef PTHREAD_SETS_ERRNO
         rv = errno;
 #endif
-#ifdef HAVE_PTHREAD_MUTEXATTR_SETROBUST_NP
+#ifdef HAVE_PTHREAD_MUTEX_ROBUST
         /* Okay, our owner died.  Let's try to make it consistent again. */
         if (rv == EOWNERDEAD) {
             pthread_mutex_consistent_np(mutex->pthread_interproc);
@@ -434,13 +400,13 @@ static apr_status_t proc_mutex_proc_pthread_release(apr_proc_mutex_t *mutex)
 {
     apr_status_t rv;
 
+    mutex->curr_locked = 0;
     if ((rv = pthread_mutex_unlock(mutex->pthread_interproc))) {
 #ifdef PTHREAD_SETS_ERRNO
         rv = errno;
 #endif
         return rv;
     }
-    mutex->curr_locked = 0;
     return APR_SUCCESS;
 }
 
@@ -456,7 +422,7 @@ const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_proc_pthread_method
     APR_PROCESS_LOCK_MECH_IS_GLOBAL,
     proc_mutex_proc_pthread_create,
     proc_mutex_proc_pthread_acquire,
-    NULL, /* no tryacquire */
+    proc_mutex_no_tryacquire,
     proc_mutex_proc_pthread_release,
     proc_mutex_proc_pthread_cleanup,
     proc_mutex_proc_pthread_child_init,
@@ -496,8 +462,9 @@ static apr_status_t proc_mutex_fcntl_cleanup(void *mutex_)
         if (status != APR_SUCCESS)
             return status;
     }
-    apr_file_close(mutex->interproc);
-    
+    if (mutex->interproc) { /* if it was opened successfully */
+        apr_file_close(mutex->interproc);
+    }
     return APR_SUCCESS;
 }    
 
@@ -515,7 +482,8 @@ static apr_status_t proc_mutex_fcntl_create(apr_proc_mutex_t *new_mutex,
     }
     else {
         new_mutex->fname = apr_pstrdup(new_mutex->pool, "/tmp/aprXXXXXX");
-        rv = apr_file_mktemp(&new_mutex->interproc, new_mutex->fname, 0,
+        rv = apr_file_mktemp(&new_mutex->interproc, new_mutex->fname,
+                             APR_CREATE | APR_WRITE | APR_EXCL,
                              new_mutex->pool);
     }
  
@@ -525,11 +493,6 @@ static apr_status_t proc_mutex_fcntl_create(apr_proc_mutex_t *new_mutex,
     }
 
     new_mutex->curr_locked = 0;
-    /* XXX currently, apr_file_mktemp() always specifies that the file should
-     *     be removed when closed; that unlink() will fail since we're 
-     *     removing it here; we want to remove it here since we don't need
-     *     it visible and we want it cleaned up if we exit catastrophically
-     */
     unlink(new_mutex->fname);
     apr_pool_cleanup_register(new_mutex->pool,
                               (void*)new_mutex,
@@ -556,13 +519,13 @@ static apr_status_t proc_mutex_fcntl_release(apr_proc_mutex_t *mutex)
 {
     int rc;
 
+    mutex->curr_locked=0;
     do {
         rc = fcntl(mutex->interproc->filedes, F_SETLKW, &proc_mutex_unlock_it);
     } while (rc < 0 && errno == EINTR);
     if (rc < 0) {
         return errno;
     }
-    mutex->curr_locked=0;
     return APR_SUCCESS;
 }
 
@@ -582,7 +545,7 @@ const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_fcntl_methods =
 #endif
     proc_mutex_fcntl_create,
     proc_mutex_fcntl_acquire,
-    NULL, /* no tryacquire */
+    proc_mutex_no_tryacquire,
     proc_mutex_fcntl_release,
     proc_mutex_fcntl_cleanup,
     proc_mutex_fcntl_child_init,
@@ -609,7 +572,9 @@ static apr_status_t proc_mutex_flock_cleanup(void *mutex_)
         if (status != APR_SUCCESS)
             return status;
     }
-    apr_file_close(mutex->interproc);
+    if (mutex->interproc) { /* if it was opened properly */
+        apr_file_close(mutex->interproc);
+    }
     unlink(mutex->fname);
     return APR_SUCCESS;
 }    
@@ -628,7 +593,8 @@ static apr_status_t proc_mutex_flock_create(apr_proc_mutex_t *new_mutex,
     }
     else {
         new_mutex->fname = apr_pstrdup(new_mutex->pool, "/tmp/aprXXXXXX");
-        rv = apr_file_mktemp(&new_mutex->interproc, new_mutex->fname, 0,
+        rv = apr_file_mktemp(&new_mutex->interproc, new_mutex->fname,
+                             APR_CREATE | APR_WRITE | APR_EXCL,
                              new_mutex->pool);
     }
  
@@ -661,13 +627,13 @@ static apr_status_t proc_mutex_flock_release(apr_proc_mutex_t *mutex)
 {
     int rc;
 
+    mutex->curr_locked = 0;
     do {
         rc = flock(mutex->interproc->filedes, LOCK_UN);
     } while (rc < 0 && errno == EINTR);
     if (rc < 0) {
         return errno;
     }
-    mutex->curr_locked = 0;
     return APR_SUCCESS;
 }
 
@@ -682,6 +648,9 @@ static apr_status_t proc_mutex_flock_child_init(apr_proc_mutex_t **mutex,
 
     memcpy(new_mutex, *mutex, sizeof *new_mutex);
     new_mutex->pool = pool;
+    if (!fname) {
+        fname = (*mutex)->fname;
+    }
     new_mutex->fname = apr_pstrdup(pool, fname);
     rv = apr_file_open(&new_mutex->interproc, new_mutex->fname,
                        APR_WRITE, 0, new_mutex->pool);
@@ -702,7 +671,7 @@ const apr_proc_mutex_unix_lock_methods_t apr_proc_mutex_unix_flock_methods =
 #endif
     proc_mutex_flock_create,
     proc_mutex_flock_acquire,
-    NULL, /* no tryacquire */
+    proc_mutex_no_tryacquire,
     proc_mutex_flock_release,
     proc_mutex_flock_cleanup,
     proc_mutex_flock_child_init,
@@ -874,6 +843,18 @@ APR_DECLARE(const char *) apr_proc_mutex_name(apr_proc_mutex_t *mutex)
     return mutex->meth->name;
 }
 
+APR_DECLARE(const char *) apr_proc_mutex_lockfile(apr_proc_mutex_t *mutex)
+{
+    /* posix sems use the fname field but don't use a file,
+     * so be careful 
+     */
+    if (!strcmp(mutex->meth->name, "flock") ||
+        !strcmp(mutex->meth->name, "fcntl")) {
+        return mutex->fname;
+    }
+    return NULL;
+}
+
 APR_POOL_IMPLEMENT_ACCESSOR(proc_mutex)
 
 /* Implement OS-specific accessors defined in apr_portable.h */
